import numpy as np
from sklearn.neighbors import KNeighborsClassifier

from ._utils import *


class Metrics:
    @staticmethod
    def knn_score(embeddings: np.ndarray, labels: np.ndarray, n_neighbors=10) -> int:
        """performs KNN classifier on the spectral-embedding space and computes its accuracy.

        Parameters
        ----------
        embeddings : np.ndarray
            The spectral-embedding space.
        labels : np.ndarray
            The Spectral-embedding points' labels
        n_neighbors : int (default = 5)
            the number of neighbors for computing the KNN algorithm.

        Returns
        -------
        int
            the knn classifier accuracy for the given data.
        """

        knn_classifier = KNeighborsClassifier(n_neighbors=n_neighbors + 1, weights=knn_weights_without_self)
        knn_classifier.fit(embeddings, labels)
        knn_pred = knn_classifier.predict(embeddings)
        knn_acc = np.mean((knn_pred == labels) * 1)
        return knn_acc



import numpy as np
import torch
from sklearn.cluster import KMeans

from ._trainers import SpectralTrainer, AETrainer
from ._utils import *


class SpectralNet:
    def __init__(
            self,
            n_clusters: int,
            should_use_ae: bool = False,
            is_sparse_graph: bool = False,
            ae_hiddens: list = [512, 512, 2048, 10],
            ae_epochs: int = 40,
            ae_lr: float = 1e-3,
            ae_lr_decay: float = 0.1,
            ae_min_lr: float = 1e-7,
            ae_patience: int = 10,
            ae_batch_size: int = 256,
            spectral_hiddens: list = [1024, 1024, 512, 10],
            spectral_epochs: int = 30,
            spectral_lr: float = 1e-3,
            spectral_lr_decay: float = 0.1,
            spectral_min_lr: float = 1e-8,
            spectral_patience: int = 10,
            spectral_batch_size: int = 1024,
            spectral_n_nbg: int = 30,
            spectral_scale_k: int = 15,
            spectral_is_local_scale: bool = True,
    ):
        """SpectralNet is a class for implementing a Deep learning model that performs spectral clustering.
        This model optionally utilizes Autoencoders (AE) for training.

        Parameters
        ----------
        n_clusters : int
            The number of clusters to be generated by the SpectralNet algorithm.
            Also used for the dimention of the projection subspace.

        should_use_ae : bool, optional (default=False)
            Specifies whether to use the Autoencoder (AE) network as part of the training process.

        is_sparse_graph : bool, optional (default=False)
            Specifies whether the graph Laplacian created from the data is sparse.

        ae_hiddens : list, optional (default=[512, 512, 2048, 10])
            The number of hidden units in each layer of the Autoencoder network.

        ae_epochs : int, optional (default=30)
            The number of epochs to train the Autoencoder network.

        ae_lr : float, optional (default=1e-3)
            The learning rate for the Autoencoder network.

        ae_lr_decay : float, optional (default=0.1)
            The learning rate decay factor for the Autoencoder network.

        ae_min_lr : float, optional (default=1e-7)
            The minimum learning rate for the Autoencoder network.

        ae_patience : int, optional (default=10)
            The number of epochs to wait before reducing the learning rate for the Autoencoder network.

        ae_batch_size : int, optional (default=256)
            The batch size used during training of the Autoencoder network.

        spectral_hiddens : list, optional (default=[1024, 1024, 512, 10])
            The number of hidden units in each layer of the Spectral network.

        spectral_epochs : int, optional (default=30)
            The number of epochs to train the Spectral network.

        spectral_lr : float, optional (default=1e-3)
            The learning rate for the Spectral network.

        spectral_lr_decay : float, optional (default=0.1)
            The learning rate decay factor"""

        self.n_clusters = n_clusters
        self.should_use_ae = should_use_ae
        self.is_sparse_graph = is_sparse_graph
        self.ae_hiddens = ae_hiddens
        self.ae_epochs = ae_epochs
        self.ae_lr = ae_lr
        self.ae_lr_decay = ae_lr_decay
        self.ae_min_lr = ae_min_lr
        self.ae_patience = ae_patience
        self.ae_batch_size = ae_batch_size
        self.spectral_hiddens = spectral_hiddens
        self.spectral_epochs = spectral_epochs
        self.spectral_lr = spectral_lr
        self.spectral_lr_decay = spectral_lr_decay
        self.spectral_min_lr = spectral_min_lr
        self.spectral_patience = spectral_patience
        self.spectral_n_nbg = spectral_n_nbg
        self.spectral_scale_k = spectral_scale_k
        self.spectral_is_local_scale = spectral_is_local_scale
        self.spectral_batch_size = spectral_batch_size
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self._validate_spectral_hiddens()

    def _validate_spectral_hiddens(self):
        """Validates the number of hidden units in each layer of the Spectral network."""

        if self.spectral_hiddens[-1] != self.n_clusters+1:
            raise ValueError(
                "The number of units in the last layer of spectral_hiddens network must be equal to the number of clusters or components."
            )

    def fit(self, X: torch.Tensor, y: torch.Tensor = None):
        """Performs the main training loop for the SpectralNet model.

        Parameters
        ----------
        X : torch.Tensor
            Data to train the networks on.

        y : torch.Tensor, optional
            Labels in case there are any. Defaults to None.
        """
        self._X = X
        ae_config = {
            "hiddens": self.ae_hiddens,
            "epochs": self.ae_epochs,
            "lr": self.ae_lr,
            "lr_decay": self.ae_lr_decay,
            "min_lr": self.ae_min_lr,
            "patience": self.ae_patience,
            "batch_size": self.ae_batch_size,
        }

        spectral_config = {
            "hiddens": self.spectral_hiddens,
            "epochs": self.spectral_epochs,
            "lr": self.spectral_lr,
            "lr_decay": self.spectral_lr_decay,
            "min_lr": self.spectral_min_lr,
            "patience": self.spectral_patience,
            "n_nbg": self.spectral_n_nbg,
            "scale_k": self.spectral_scale_k,
            "is_local_scale": self.spectral_is_local_scale,
            "batch_size": self.spectral_batch_size,
        }

        if self.should_use_ae:
            self.ae_trainer = AETrainer(config=ae_config, device=self.device)
            self.ae_net = self.ae_trainer.train(X)
            X = self.ae_trainer.embed(X)

        is_sparse = self.is_sparse_graph
        if is_sparse:
            build_ann(X)

        self.spectral_trainer = SpectralTrainer(
            config=spectral_config, device=self.device, is_sparse=is_sparse
        )
        self.spec_net = self.spectral_trainer.train(X, y)

    def predict(self, X: torch.Tensor):
        """Predicts the cluster assignments for the given data.

        Parameters
        ----------
        X : torch.Tensor
            Data to be clustered.

        Returns
        -------
        """
        X = X.view(X.size(0), -1)
        X = X.to(self.device)

        with torch.no_grad():
            if self.should_use_ae:
                X = self.ae_net.encode(X)
            self.embeddings_ = self.spec_net(X, should_update_orth_weights=False)
            self.embeddings_ = self.embeddings_.detach().cpu().numpy()

    def get_random_batch(self, batch_size: int = 1024) -> tuple:
        """Get a batch of the input data.

        Parameters
        ----------
        batch_size : int
            The size of the batch to use.

        Returns
        -------
        tuple
            The raw batch and the encoded batch.

        """
        permuted_indices = torch.randperm(batch_size)
        X_raw = self._X.view(self._X.size(0), -1)
        X_encoded = X_raw

        if self.should_use_ae:
            X_encoded = self.ae_trainer.embed(self._X)

        X_encoded = X_encoded[permuted_indices]
        X_raw = X_raw[permuted_indices]
        X_encoded = X_encoded.to(self.device)
        return X_raw, X_encoded

    def _get_clusters_by_kmeans(self, embeddings: np.ndarray) -> np.ndarray:
        """Performs k-means clustering on the spectral-embedding space.

        Parameters
        ----------
        embeddings : np.ndarray
            The spectral-embedding space.

        Returns
        -------
        np.ndarray
            The cluster assignments for the given data.
        """

        kmeans = KMeans(n_clusters=self.n_clusters, n_init=10).fit(embeddings)
        cluster_assignments = kmeans.predict(embeddings)
        return cluster_assignments



import numpy as np
import torch
from scipy.sparse.linalg import eigsh

from ._cluster import SpectralNet
from ._utils import *


class GrEASE:
    def __init__(
            self,
            n_components: int = None,
            is_sparse_graph: bool = False,
            spectral_hiddens: list = [1024, 1024, 512],
            spectral_max_epochs: int = 200,
            spectral_lr: float = 1e-2,
            spectral_lr_decay: float = 0.1,
            spectral_min_lr: float = 1e-6,
            spectral_patience: int = None,
            spectral_batch_size: int = 2048,
            spectral_n_nbg: int = None,
            spectral_scale_k: int = 15,
            spectral_is_local_scale: bool = True,
            should_true_eigenvectors: bool = True,
            should_return_first_eigenvector: bool = False,
            t: int = 0,
            ae_hiddens: list = [512, 512, 2048, 10],
            ae_epochs: int = 40,
            ae_lr: float = 1e-3,
            ae_lr_decay: float = 0.1,
            ae_min_lr: float = 1e-7,
            ae_patience: int = 10,
            ae_batch_size: int = 256,
            should_use_ae: bool = False,
            laplacian_kind: str = 'rw',
    ):
        """SpectralNet is a class for implementing a Deep learning model that performs spectral clustering.
        This model optionally utilizes Autoencoders (AE) for training.

        Parameters
        ----------
        n_components : int (default=None)
            The number of components to keep.

        should_use_ae : bool, optional (default=False)
            Specifies whether to use the Autoencoder (AE) network as part of the training process.

        is_sparse_graph : bool, optional (default=False)
            Specifies whether the graph Laplacian created from the data is sparse.

        ae_hiddens : list, optional (default=[512, 512, 2048, 10])
            The number of hidden units in each layer of the Autoencoder network.

        ae_epochs : int, optional (default=30)
            The number of epochs to train the Autoencoder network.

        ae_lr : float, optional (default=1e-3)
            The learning rate for the Autoencoder network.

        ae_lr_decay : float, optional (default=0.1)
            The learning rate decay factor for the Autoencoder network.

        ae_min_lr : float, optional (default=1e-7)
            The minimum learning rate for the Autoencoder network.

        ae_patience : int, optional (default=10)
            The number of epochs to wait before reducing the learning rate for the Autoencoder network.

        ae_batch_size : int, optional (default=256)
            The batch size used during training of the Autoencoder network.

        spectral_hiddens : list, optional (default=[1024, 1024, 512, 10])
            The number of hidden units in each layer of the Spectral network (not including the output layer).

        spectral_max_epochs : int, optional (default=30)
            The number of epochs to train the Spectral network.

        spectral_lr : float, optional (default=1e-3)
            The learning rate for the Spectral network.

        spectral_lr_decay : float, optional (default=0.1)
            The learning rate decay factor.

        should_true_eigenvectors : bool, optional (default=True)
            Specifies whether to compute the true eigenvectors of the Laplacian of the input data.

        should_return_first_eigenvector : bool, optional (default=False)
            Specifies whether to return the first eigenvector of the Laplacian of the input data.

        t : int, optional (default=0)
            The diffusion time for the diffusion map algorithm.

        laplacian_kind : str, optional (default='rw')
            Specifies the kind of Laplacian matrix to use. It can be one of 'unnormalized', 'rw', or 'symmetric'."""

        self.n_components = n_components
        self.is_sparse_graph = is_sparse_graph
        self.spectral_hiddens = spectral_hiddens
        self.spectral_hiddens.append(self.n_components + 1)
        self.spectral_epochs = spectral_max_epochs
        self.spectral_lr = spectral_lr
        self.spectral_lr_decay = spectral_lr_decay
        self.spectral_min_lr = spectral_min_lr
        self.spectral_patience = spectral_patience
        self.spectral_batch_size = spectral_batch_size
        self.spectral_n_nbg = spectral_n_nbg
        if spectral_n_nbg is None:
            self.spectral_n_nbg = max(5, self.spectral_batch_size // 200)
        self.spectral_scale_k = spectral_scale_k
        self.spectral_is_local_scale = spectral_is_local_scale
        self.X_new = None
        self.ortho_matrix = np.eye(n_components+1)
        self.eigenvalues = np.ones(n_components+1)
        self.t = t
        self.columns = None
        if t > 0:
            self.should_true_eigenvectors = True
            self.should_return_first_eigenvector = False
        else:
            self.should_true_eigenvectors = should_true_eigenvectors
            self.should_return_first_eigenvector = should_return_first_eigenvector
        self.laplacian_kind = laplacian_kind

        # AE
        self.ae_hiddens = ae_hiddens
        self.ae_epochs = ae_epochs
        self.ae_lr = ae_lr
        self.ae_lr_decay = ae_lr_decay
        self.ae_min_lr = ae_min_lr
        self.ae_patience = ae_patience
        self.ae_batch_size = ae_batch_size
        self.should_use_ae = should_use_ae

    def fit(self, X: torch.Tensor, y: torch.Tensor = None):
        """Fit the SpectralNet model to the input data.

        Parameters
        ----------
        X : torch.Tensor
            The input data of shape (n_samples, n_features).

        y: torch.Tensor
            The labels of the input data of shape (n_samples,).

        Returns
        -------
        np.ndarray
            The fitted embeddings of shape (n_samples, n_components).
        """
        X = normalize_data(X)

        if self.n_components is None:
            raise ValueError("The number of components must be specified.")

        self.spectral_batch_size = min(self.spectral_batch_size, X.shape[0])

        # Set the spectral patience if it is not set
        if self.spectral_patience is None:
            n_batches = (X.shape[0] // self.spectral_batch_size)
            if n_batches <= 25:
                self.spectral_patience = 10
            else:
                self.spectral_patience = max(1, 250 // n_batches)

        self._spectralnet = SpectralNet(
            n_clusters=self.n_components,
            spectral_hiddens=self.spectral_hiddens,
            spectral_epochs=self.spectral_epochs,
            spectral_lr=self.spectral_lr,
            spectral_lr_decay=self.spectral_lr_decay,
            spectral_min_lr=self.spectral_min_lr,
            spectral_patience=self.spectral_patience,
            spectral_n_nbg=self.spectral_n_nbg,
            spectral_scale_k=self.spectral_scale_k,
            spectral_is_local_scale=self.spectral_is_local_scale,
            spectral_batch_size=self.spectral_batch_size,
            ae_hiddens=self.ae_hiddens,
            ae_epochs=self.ae_epochs,
            ae_lr=self.ae_lr,
            ae_lr_decay=self.ae_lr_decay,
            ae_min_lr=self.ae_min_lr,
            ae_patience=self.ae_patience,
            ae_batch_size=self.ae_batch_size,
            should_use_ae=self.should_use_ae,
            laplacian_kind=self.laplacian_kind,
        )

        self._spectralnet.fit(X, y)

        if self.should_true_eigenvectors:
            self.compute_ortho_matrix(X)

    def _predict(self, X: torch.Tensor) -> np.ndarray:
        """Predict embeddings for the input data using the fitted SpectralNet model.

        Parameters
        ----------
        X : torch.Tensor
            The input data of shape (n_samples, n_features).

        Returns
        -------
        np.ndarray
            The predicted embeddings of shape (n_samples, n_components).
        """
        self._spectralnet.predict(X)
        return self._spectralnet.embeddings_

    def transform(self, X: torch.Tensor) -> np.ndarray:
        """Transform the input data into embeddings using the fitted SpectralNet model.

        Parameters
        ----------
        X : torch.Tensor
            The input data of shape (n_samples, n_features).

        Returns
        -------
        np.ndarray
            The transformed embeddings of shape (n_samples, n_components).

        or

        tuple
            The transformed embeddings of shape (n_samples, n_components) and the eigenvalues of the Laplacian of the input data.
        """
        X = normalize_data(X)
        if not self.should_true_eigenvectors:
            return self._predict(X)
        if self.should_return_first_eigenvector:
            return self._predict(X) @ self.ortho_matrix
        return (self._predict(X) @ self.ortho_matrix @ np.diag((1 - self.eigenvalues) ** self.t))[:, 1:]

    def fit_transform(self, X: torch.Tensor, y: torch.Tensor = None) -> np.ndarray:
        """Fit the SpectralNet model to the input data and transform it into embeddings.

        This is a convenience method that combines the fit and transform steps.

        Parameters
        ----------
        X : torch.Tensor
            The input data of shape (n_samples, n_features).

        y: torch.Tensor
            The labels of the input data of shape (n_samples,).

        Returns
        -------
        np.ndarray
            The fitted and transformed embeddings of shape (n_samples, n_components).
        """
        self.fit(X, y)
        return self.transform(X)

    def get_eigenvalues(self):
        return self.eigenvalues[1:]

    def _get_laplacian_of_small_batch(self, batch: torch.Tensor) -> np.ndarray:
        """Get the Laplacian of a small batch of the input data

        Parameters
        ----------

        batch : torch.Tensor
            A small batch of the input data of shape (batch_size, n_features).

        Returns
        -------
        np.ndarray
            The Laplacian of the small batch of the input data.



        """

        W = get_affinity_matrix(batch, n_nbg=self.spectral_n_nbg, device=self._spectralnet.device)
        L = get_laplacian(W)
        return L

    def compute_ortho_matrix(self, X: torch.Tensor) -> None:
        """Compute the orthogonal matrix for the spectral embeddings.

        Parameters
        ----------
        X : torch.Tensor
            The input data of shape (n_samples, n_features).
        """
        pred = self._predict(X)
        Lambda = self._get_lambda_on_multi_batches(X)

        try:
            ortho_matrix, eigenvalues_pred, _ = np.linalg.svd(Lambda)
            eigenvalues_pred = eigenvalues_pred.real
            self.ortho_matrix = np.array(ortho_matrix.real)
            indices = np.argsort(eigenvalues_pred)
            self.ortho_matrix = np.array(self.ortho_matrix[:, indices])
            self.eigenvalues = eigenvalues_pred[indices]
        except np.linalg.LinAlgError:
            print("Warning: SVD did not converge")
            self.eigenvalues = np.diag(np.sort(Lambda))

    def _get_lambda_on_multi_batches(self, X) -> np.ndarray:
        """Get the mean eigenvalues matrix of the Laplacian of the input data.

        Returns
        -------
        np.ndarray
            The eigenvalues matrix of the Laplacian of the input data.
        """
        n_batches = (X.shape[0] // self.spectral_batch_size)
        Lambda = self._get_lambda_on_batch(self.spectral_batch_size)
        for i in range(1, n_batches):
            Lambda += self._get_lambda_on_batch()
        Lambda = torch.Tensor(Lambda) / n_batches
        return (Lambda + Lambda.T) / 2

    def _get_lambda_on_batch(self, batch_size=None) -> np.ndarray:
        """Get the eigenvalues of the Laplacian of a small batch of the input data.

        Returns
        -------
        np.ndarray
            The eigenvalues of the Laplacian of a small batch of the input data.
        """

        if batch_size is None:
            batch_size = self.spectral_batch_size
        batch_raw, batch_encoded = self._spectralnet.get_random_batch(batch_size)
        L_batch = self._get_laplacian_of_small_batch(batch_encoded)
        V_batch = self._predict(batch_raw)
        V_batch = V_batch / np.linalg.norm(V_batch, axis=0)
        Lambda = V_batch.T @ L_batch @ V_batch
        return Lambda

    def grassmann_distance(self, X: torch.Tensor, each_vector: bool = False) -> float:
        """Compute the Grassmann distance between the embeddings of the true SE"""
        pred = self.transform(X)
        se = self._get_true_se(X)
        return get_grassmann_distance(pred, se, each_vector)

    def _get_true_se(self, X: torch.Tensor) -> np.ndarray:
        """Get the true spectral embeddings of the input data.

        Parameters
        ----------
        X : torch.Tensor
            The input data of shape (n_samples, n_features).

        Returns
        -------
        np.ndarray
            The true spectral embeddings of the input data.
        """
        n_neighbors = (X.shape[0] // self.spectral_batch_size) * self.spectral_n_nbg
        W = get_affinity_matrix(X, n_nbg=n_neighbors, device=self._spectralnet.device)
        L = get_laplacian(W, laplacian_kind=self.laplacian_kind)
        V = eigsh(L, k=self.n_components+1, which='SM')[1]
        return V[:, 1:]



import torch
import torch.nn as nn


class SpectralNetLoss(nn.Module):
    def __init__(self):
        super(SpectralNetLoss, self).__init__()

    def forward(
        self, W: torch.Tensor, Y: torch.Tensor, is_normalized: bool = True
    ) -> torch.Tensor:
        """
        This function computes the loss of the SpectralNet model.
        The loss is the rayleigh quotient of the Laplacian matrix obtained from W,
        and the orthonormalized output of the network.

        Args:
            W (torch.Tensor):               Affinity matrix
            Y (torch.Tensor):               Output of the network
            is_normalized (bool, optional): Whether to use the RW-normalized Laplacian matrix or not.

        Returns:
            torch.Tensor: The loss
        """
        m = Y.size(0)
        if is_normalized:
            D = torch.sum(W, dim=1)
            Y = Y / torch.sqrt(D)[:, None]

        Dy = torch.cdist(Y, Y)
        loss = torch.sum(W * Dy.pow(2)) / (2 * m)

        return loss



import torch
import torch.optim as optim
from torch.utils.data import DataLoader, random_split, TensorDataset
from tqdm import trange

from .._utils import *
from .._losses import SpectralNetLoss
from .._models import SpectralNetModel


class SpectralTrainer:
    def __init__(self, config: dict, device: torch.device, is_sparse: bool = False):
        """
        Initialize the SpectralNet model trainer.

        Parameters
        ----------
        config : dict
            The configuration dictionary.
        device : torch.device
            The device to use for training.
        is_sparse : bool, optional
            Whether the graph-laplacian obtained from a mini-batch is sparse or not.
            If True, the batch is constructed by taking 1/5 of the original random batch
            and adding 4 of its nearest neighbors to each sample. Defaults to False.

        Notes
        -----
        This class is responsible for training the SpectralNet model.
        The configuration dictionary (`config`) contains various settings for training.
        The device (`device`) specifies the device (CPU or GPU) to be used for training.
        The `is_sparse` flag is used to determine the construction of the batch when the graph-laplacian is sparse.
        """

        self.device = device
        self.is_sparse = is_sparse
        self.spectral_config = config
        self.lr = self.spectral_config["lr"]
        self.n_nbg = self.spectral_config["n_nbg"]
        self.min_lr = self.spectral_config["min_lr"]
        self.epochs = self.spectral_config["epochs"]
        self.scale_k = self.spectral_config["scale_k"]
        self.lr_decay = self.spectral_config["lr_decay"]
        self.patience = self.spectral_config["patience"]
        self.architecture = self.spectral_config["hiddens"]
        self.batch_size = self.spectral_config["batch_size"]
        self.is_local_scale = self.spectral_config["is_local_scale"]
        self.laplacian_kind = self.spectral_config["laplacian_kind"]

    def train(
            self, X: torch.Tensor, y: torch.Tensor
    ) -> SpectralNetModel:
        """
        Train the SpectralNet model.

        Parameters
        ----------
        X : torch.Tensor
            The dataset to train on.
        y : torch.Tensor, optional
            The labels of the dataset in case there are any.

        Returns
        -------
        SpectralNetModel
            The trained SpectralNet model.

        Notes
        -----
        This function trains the SpectralNet model using the provided dataset (`X`) and labels (`y`).
        If labels are not provided (`y` is None), unsupervised training is performed.
        The trained SpectralNet model is returned as the output.
        """

        self.X = X.view(X.size(0), -1)
        self.y = y
        self.counter = 0
        self.criterion = SpectralNetLoss(self.laplacian_kind)
        self.spectral_net = SpectralNetModel(
            self.architecture, input_dim=self.X.shape[1]
        ).to(self.device)

        self.optimizer = optim.Adam(self.spectral_net.parameters(), lr=self.lr)

        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, mode="min", factor=self.lr_decay, patience=self.patience
        )

        train_loader, ortho_loader, valid_loader = self._get_data_loader()

        print("Training SpectralNet:")
        t = trange(self.epochs, leave=True)
        for epoch in t:
            train_loss = 0.0
            for (X_grad, _), (X_orth, _) in zip(train_loader, ortho_loader):
                X_grad = X_grad.to(device=self.device)
                X_grad = X_grad.view(X_grad.size(0), -1)
                X_orth = X_orth.to(device=self.device)
                X_orth = X_orth.view(X_orth.size(0), -1)

                if self.is_sparse:
                    X_grad = make_batch_for_sparse_grapsh(X_grad)
                    X_orth = make_batch_for_sparse_grapsh(X_orth)

                # Orthogonalization step
                self.spectral_net.eval()
                self.spectral_net(X_orth, should_update_orth_weights=True)

                # Gradient step
                self.spectral_net.train()
                self.optimizer.zero_grad()

                Y = self.spectral_net(X_grad, should_update_orth_weights=False)

                W = self._get_affinity_matrix(X_grad)

                loss = self.criterion(W, Y)
                loss.backward()
                self.optimizer.step()
                train_loss += loss.item()

            train_loss /= len(train_loader)

            # Validation step
            valid_loss = self.validate(valid_loader)
            self.scheduler.step(valid_loss)

            current_lr = self.optimizer.param_groups[0]["lr"]
            if current_lr <= self.spectral_config["min_lr"]:
                break
            t.set_description(
                "Train Loss: {:.7f}, Valid Loss: {:.7f}, LR: {:.6f}".format(
                    train_loss, valid_loss, current_lr
                )
            )
            t.refresh()

        return self.spectral_net

    def validate(self, valid_loader: DataLoader) -> float:
        valid_loss = 0.0
        self.spectral_net.eval()
        with torch.no_grad():
            for batch in valid_loader:
                X, y = batch
                X, y = X.to(self.device), y.to(self.device)

                if self.is_sparse:
                    X = make_batch_for_sparse_grapsh(X)

                Y = self.spectral_net(X, should_update_orth_weights=False)

                W = self._get_affinity_matrix(X)

                loss = self.criterion(W, Y)
                valid_loss += loss.item()

        valid_loss /= len(valid_loader)

        return valid_loss

    def _get_affinity_matrix(self, X: torch.Tensor) -> torch.Tensor:
        """
        This function computes the affinity matrix W using the Gaussian kernel.

        Args:
            X (torch.Tensor):   The input data

        Returns:
            torch.Tensor: The affinity matrix W
        """

        is_local = self.is_local_scale
        n_neighbors = self.n_nbg
        scale_k = self.scale_k
        Dx = torch.cdist(X, X)
        Dis, indices = get_nearest_neighbors(X, k=n_neighbors + 1)
        Dis -= Dis[:, 0].reshape(Dis.shape[0], 1)
        Dis[Dis < 0] = 0
        scale = compute_scale(Dis, k=scale_k, is_local=is_local)
        W = get_laplace_kernel(
            Dx, scale, indices[:, 1:], device=self.device, is_local=is_local
        )
        return W

    def _get_data_loader(self) -> tuple:
        """
        This function returns the data loaders for training, validation and testing.

        Returns:
            tuple:  The data loaders
        """
        if self.y is None:
            self.y = torch.zeros(len(self.X))
        train_size = int(0.9 * len(self.X))
        valid_size = len(self.X) - train_size
        dataset = TensorDataset(self.X, self.y)
        train_dataset, valid_dataset = random_split(dataset, [train_size, valid_size])
        train_loader = DataLoader(
            train_dataset, batch_size=self.batch_size, shuffle=True
        )
        ortho_loader = DataLoader(
            train_dataset, batch_size=self.batch_size, shuffle=True
        )
        valid_loader = DataLoader(
            valid_dataset, batch_size=self.batch_size, shuffle=False
        )
        return train_loader, ortho_loader, valid_loader



import torch
import torch.nn as nn


class SpectralNetLoss(nn.Module):
    def __init__(self, laplacian_kind: str='rw'):
        super(SpectralNetLoss, self).__init__()
        self.laplacian_kind = laplacian_kind

    def forward(
        self, W: torch.Tensor, Y: torch.Tensor
    ) -> torch.Tensor:
        """
        This function computes the loss of the SpectralNet model.
        The loss is the rayleigh quotient of the Laplacian matrix obtained from W,
        and the orthonormalized output of the network.

        Args:
            W (torch.Tensor):               Affinity matrix
            Y (torch.Tensor):               Output of the network
            laplacian_kind (bool, optional): Specifies the kind of Laplacian matrix to use. Defaults to RW.

        Returns:
            torch.Tensor: The loss
        """
        m = Y.size(0)
        # if self.is_normalized:
        #     D = torch.sum(W, dim=1)
        #     Y = Y / torch.sqrt(D)[:, None]

        # Dy = torch.cdist(Y, Y)
        # loss = torch.sum(W * Dy.pow(2)) / (2 * m)

        D = torch.diag(torch.sum(W, dim=1)).to(W.device)
        if self.laplacian_kind == "unnormalized":
            L = D - W
        elif self.laplacian_kind == "rw":
            L = torch.eye(m).to(W.device) - torch.inverse(D) @ W
        else: # self.laplacian_kind == "symmetric"
            D_inv_sqrt = torch.sqrt(torch.inverse(D))
            L = torch.eye(m).to(W.device) - D_inv_sqrt @ W @ D_inv_sqrt

        loss = torch.trace(Y.T @ L @ Y) / m

        return loss



import os
import torch
import torch.nn as nn
import torch.optim as optim

from tqdm import trange
from ._trainer import Trainer
from .._models import AEModel
from torch.utils.data import DataLoader, random_split


class AETrainer:
    def __init__(self, config: dict, device: torch.device):
        self.device = device
        self.ae_config = config
        self.lr = self.ae_config["lr"]
        self.epochs = self.ae_config["epochs"]
        self.min_lr = self.ae_config["min_lr"]
        self.lr_decay = self.ae_config["lr_decay"]
        self.patience = self.ae_config["patience"]
        self.architecture = self.ae_config["hiddens"]
        self.batch_size = self.ae_config["batch_size"]
        self.weights_dir = "spectralnet/_trainers/weights"
        self.weights_path = "spectralnet/_trainers/weights/ae_weights.pth"
        if not os.path.exists(self.weights_dir):
            os.makedirs(self.weights_dir)

    def train(self, X: torch.Tensor) -> AEModel:
        self.X = X.view(X.size(0), -1)
        self.criterion = nn.MSELoss()

        self.ae_net = AEModel(self.architecture, input_dim=self.X.shape[1]).to(
            self.device
        )

        self.optimizer = optim.Adam(self.ae_net.parameters(), lr=self.lr)

        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, mode="min", factor=self.lr_decay, patience=self.patience
        )

        if os.path.exists(self.weights_path):
            self.ae_net.load_state_dict(torch.load(self.weights_path))
            return self.ae_net

        train_loader, valid_loader = self._get_data_loader()

        print("Training Autoencoder:")
        t = trange(self.epochs, leave=True)
        for epoch in t:
            train_loss = 0.0
            for batch_x in train_loader:
                batch_x = batch_x.to(self.device)
                batch_x = batch_x.view(batch_x.size(0), -1)
                self.optimizer.zero_grad()
                output = self.ae_net(batch_x)
                loss = self.criterion(output, batch_x)
                loss.backward()
                self.optimizer.step()
                train_loss += loss.item()

            train_loss /= len(train_loader)
            valid_loss = self.validate(valid_loader)
            self.scheduler.step(valid_loss)
            current_lr = self.optimizer.param_groups[0]["lr"]

            if current_lr <= self.min_lr:
                break

            t.set_description(
                "Train Loss: {:.7f}, Valid Loss: {:.7f}, LR: {:.6f}".format(
                    train_loss, valid_loss, current_lr
                )
            )
            t.refresh()

        torch.save(self.ae_net.state_dict(), self.weights_path)
        return self.ae_net

    def validate(self, valid_loader: DataLoader) -> float:
        self.ae_net.eval()
        valid_loss = 0.0
        with torch.no_grad():
            for batch_x in valid_loader:
                batch_x = batch_x.to(self.device)
                batch_x = batch_x.view(batch_x.size(0), -1)
                output = self.ae_net(batch_x)
                loss = self.criterion(output, batch_x)
                valid_loss += loss.item()
        valid_loss /= len(valid_loader)
        return valid_loss

    def embed(self, X: torch.Tensor) -> torch.Tensor:
        print("Embedding data ...")
        self.ae_net.eval()
        with torch.no_grad():
            X = X.view(X.size(0), -1)
            encoded_data = self.ae_net.encode(X.to(self.device))
        return encoded_data

    def _get_data_loader(self) -> tuple:
        trainset_len = int(len(self.X) * 0.9)
        validset_len = len(self.X) - trainset_len
        trainset, validset = random_split(self.X, [trainset_len, validset_len])
        train_loader = DataLoader(
            trainset, batch_size=self.ae_config["batch_size"], shuffle=True
        )
        valid_loader = DataLoader(
            validset, batch_size=self.ae_config["batch_size"], shuffle=False
        )
        return train_loader, valid_loader


from setuptools import setup, find_packages

setup(
    name="grease-embeddings",
    version="0.1.5",
    description="Generalizable and Efficient Approximate Spectral Embeddings",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Nir Ben-Ari",
    author_email="nirnirba@gmail.com",
    url="https://github.com/TheNirnir/GrEASE",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.9",
)
